# mqttclient.py
# -*- coding: utf-8 -*-

from threading import Thread
from paho.mqtt.client import Client, MQTT_ERR_SUCCESS
import thread

MQTTCLIENT_VERSION = "1.12 - Jan. 17, 2018"

class _MessageThread(Thread):
    def __init__(self, client, messageReceived, topic, payload):
        Thread.__init__(self)
        self.client = client
        self.messageReceived = messageReceived
        self.topic = topic
        self.payload = payload
        self.start()
        
    def run(self):
        try:
            self.messageReceived(self.topic, self.payload)
        except Exception, e:        
            self.client._debug("Exception in message callback: " + str(e))

class MQTTClient(Thread):
    '''
    Class that represents a MQTT client based on paho-mqtt package.
    '''
    def __init__(self, messageReceived = None, username = "", password = ""):
        '''
        Creates a MQTTClient that publishes and/or subcribes MQTT topics. Messages from subcribed topics trigger the 
        given messageReceived function. The client does not yet connect to a MQTT broker.
        @param messageReceived: the callback function with parameters topic, msg triggerd by incoming messages
        running in a separate thread. If None, no message notifications are triggered, e.g. for a client that only publishes topics.
        @param username: the username used to log into the MQTT broker (empty, if no user authentication is necessary)
        @param password: the password used to log into the MQTT broker (empty, if no user authentication is necessary)
        '''
        Thread.__init__(self)
        self.username = username
        self.password = password
        self.messageReceived = messageReceived
        self.verbose = False
        self.log = False
        self.client = Client()
        self.client.on_message = self._onMessage
        self.client.on_log = self._onLog
        if username != "":
            self.client.username_pw_set(username, password)
            
    def run(self):            
        self._debug("MQTTHander thread started")
        self.client.loop_forever()
        self._debug("MQTTHandler thread terminated")
        
    def setVerbose(self, verbose):
        '''
        Enables/disables logger messages printed to stdout.
        @param verbose: if True, the logger messages are enabled; otherwise disabled
        '''    
        self.verbose = verbose

    def connect(self, host, port = 1883, timeout = 60):
        '''
        Starts a connection trial to the given MQTT broker at given port (default: 1883).
        @host: the IP address if the broker
        @port: the port where the broker is listening (default: 1883)
        @param keepalive: maximum period in seconds between communications with the broker. 
        If no other messages are exchanged, time period for ping messages to the broker (default: 60 s)
        '''
        self.host = host
        self.port = port
        self._debug("connect() with host: " + host + " port: " + str(port) + " timeout: " + str(timeout))
        try:
            self.client.connect(host, port, timeout)
        except:
            return False
        if self.messageReceived != None:
            self.start() # start thread only if necessary get message callbacks
        return True

    def disconnect(self):
        '''
        Disconnects the client from the broker.
        '''
        self._debug("disconnect()")
        self.client.disconnect()

    def publish(self, topic, payload, qos = 0, retain = False):
        '''
        Sends a message with given topic and payload to the broker.
        @param topic: the topic that the message should be published on (string)
        @param payload: the message to send. An int or float will converted it to a string. 
        @param qos: the quality of service level to use (number 0, 1, 2) (default: 0)
        @param retain: if True, the message is the “last known good”/retained message for the topic (default: False)
        '''
        self._debug("publish() with topic: '" + topic + "' payload: " + str(payload) + " qos: " + str(qos))
        infot = self.client.publish(topic, payload, qos, retain)
        if infot.rc != MQTT_ERR_SUCCESS:
            return False
        try:
            infot.wait_for_publish()
        except:
            self._debug("wait_for_publish() throwed exception")
            return False
        return True

    def subscribe(self, topic, qos = 0):
        '''
        Subscribe the client to one or more topics.
        @param topic: a string or a list of tuples of format (topic, qos)
        @param qos: the quality of service level (number 0, 1, 2) (default: 0); not used, if topic is a list of tuples
        '''
        self._debug("subscribe() with topic: '" + topic + "' qos: " + str(qos))
        self.client.subscribe(topic, qos)
        
        def unsubscribe(self, topic):
            '''
            Unsubscribes the client for one or more topics.
                topic
                    a single string or a list of strings
            '''
            self.client.unsubscribe(topic)
        
    # ------------------- Callbacks -------------------------------------------
    def _onMessage(self, client, obj, msg):
        self._debug("onMessage() topic: '" + msg.topic + "' qos: " + str(msg.qos) + " payload: " + msg.payload)
        if self.messageReceived != None:
            _MessageThread(self, self.messageReceived, msg.topic, msg.payload)

    def _onLog(self, client, obj, level, msg):
        self._debug("onLog() msg: " + msg)

    def _debug(self, msg):
        if self.verbose:
            print("   MQTTClient-> " + msg)

